// Copyright (c) 2005 - 2017 Settlers Freaks (sf-team at siedler25.org)
//
// This file is part of Return To The Roots.
//
// Return To The Roots is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// Return To The Roots is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Return To The Roots. If not, see <http://www.gnu.org/licenses/>.
#ifndef ROADSEGMENT_H_INCLUDED
#define ROADSEGMENT_H_INCLUDED

#pragma once

#include "GameObject.h"
#include "gameTypes/Direction.h"
#include <boost/array.hpp>
#include <vector>

class nofCarrier;
class noRoadNode;
class noFlag;
class noFigure;
class SerializedGameData;

class RoadSegment : public GameObject
{
public:
    enum RoadType
    {
        RT_NORMAL, /// Normale Straße bzw. Bergstraße
        RT_DONKEY, /// Eselstraße
        RT_BOAT    /// Wasserstraße
    };

public:
    RoadSegment(const RoadType rt, noRoadNode* const f1, noRoadNode* const f2, const std::vector<Direction>& route);
    RoadSegment(SerializedGameData& sgd, const unsigned obj_id);

    /// zerstört das Objekt.
    void Destroy() override { Destroy_RoadSegment(); }
    /// serialisiert das Objekt.
    void Serialize(SerializedGameData& sgd) const override { Serialize_RoadSegment(sgd); }
    /// liefert den GO-Type.
    GO_Type GetGOT() const override { return GOT_ROADSEGMENT; }
    /// Gibt die ID (0 oder 1) eines RoadNodes dieser Straße zurück (die Flagge muss zu dieser Straße gehören, sonst kommt Müll raus!!)
    bool GetNodeID(const noRoadNode& rn) const;
    /// Gibt Straßen-Typ zurück
    RoadType GetRoadType() const { return rt; }
    /// Gibt die Länge der Staße zurück
    unsigned GetLength() const { return route.size(); }
    /// gibt Flagge 1 zurück
    noRoadNode* GetF1() const { return f1; }
    /// setzt Flagge 1 auf o
    void SetF1(noRoadNode* o) { f1 = o; }
    /// gibt Flagge 2 zurück
    noRoadNode* GetF2() const { return f2; }
    /// setzt Flagge 2 auf o
    void SetF2(noRoadNode* o) { f2 = o; }
    /// gibt die Route nr zurück
    Direction GetRoute(unsigned nr) const { return route.at(nr); }
    /// setzt die Route nr auf r
    void SetRoute(unsigned short nr, Direction r) { route[nr] = r; }
    /// gibt den Carrier nr zurück
    nofCarrier* getCarrier(unsigned char nr) const { return carriers_[nr]; }
    /// setzt den Carrier nr auf c
    void setCarrier(unsigned char nr, nofCarrier* c)
    {
        RTTR_Assert(!c || !hasCarrier(nr));
        carriers_[nr] = c;
    }
    /// haben wir den Carrier "nr"?
    bool hasCarrier(unsigned char nr) const { return (carriers_[nr] != NULL); }
    /// Braucht die Straße einen Esel? Nur wenn sie auch einen Träger schon hat!
    bool NeedDonkey() const { return (rt == RT_DONKEY && carriers_[0] && !carriers_[1]); }
    /// Hat einen Esel als Arbeiter dazubekommen.
    void GotDonkey(nofCarrier* donkey)
    {
        RTTR_Assert(!carriers_[1]);
        carriers_[1] = donkey;
    }

    /// haben wir überhaupt Carrier?
    bool isOccupied() const { return (carriers_[0] || carriers_[1]); }

    Direction GetDir(bool bwdDir, unsigned id) const
    {
        if(bwdDir)
            return route[route.size() - id - 1] + 3u;
        else
            return route[id];
    }

    /// zerteilt die Straße in 2 Teile.
    void SplitRoad(noFlag* splitflag);
    /// Überprüft ob es an den Flaggen noch Waren zu tragen gibt für den Träger.
    bool AreWareJobs(const bool flag, unsigned carrier_type, const bool take_ware_immediately) const;
    /// Eine Ware sagt Bescheid, dass sie über dem Weg getragen werden will.
    void AddWareJob(const noRoadNode* rn);
    /// Eine Ware will nicht mehr befördert werden.
    void WareJobRemoved(const noFigure* const exception);
    /// Baut die Straße zu einer Eselstraße aus.
    void UpgradeDonkeyRoad();
    /// Soll versuchen einen Esel zu bekommen.
    void TryGetDonkey();
    /// Ein Träger muss kündigen, aus welchen Gründen auch immer.
    void CarrierAbrogated(nofCarrier* carrier);
    /// given a flag returns the flag at the other end
    const noFlag& GetOtherFlag(const noFlag& flag) const;
    /// given a flag returns last direction of the route towards the other flag
    Direction GetOtherFlagDir(const noFlag& flag) const;

protected:
    /// zerstört das Objekt.
    void Destroy_RoadSegment();
    /// serialisiert das Objekt.
    void Serialize_RoadSegment(SerializedGameData& sgd) const;

private:
    /// Straßentyp
    RoadType rt;
    /// die 2 Roadnodes, die den Weg eingrenzen
    noRoadNode *f1, *f2;
    /// Beschreibung des Weges, ist length groß und liegt als Beschreibung der einzelnen Richtungen vor (von f1 zu f2)
    std::vector<Direction> route;
    /// Träger (und ggf. Esel), der auf diesem Weg arbeitet
    boost::array<nofCarrier*, 2> carriers_;
};

#endif // !ROADSEGMENT_H_INCLUDED
